/* floating status bar */
.floating-bar {
  background: var(--popover-bg-color);
  color: var(--popover-fg-color);
  padding: 4px 12px;
  margin: 4px;
  border-radius: 8px;
}

/* incognito mode */
.incognito-mode .top-bar.raised,
.incognito-mode .bottom-bar.raised {
  background-color: #eaf2fc;
}

.incognito-mode .top-bar.raised:backdrop,
.incognito-mode .bottom-bar.raised:backdrop {
  background-color: #eaf0f7;
}

@media (prefers-color-scheme: dark) {
  /* dark style */
  .incognito-mode .top-bar.raised,
  .incognito-mode .bottom-bar.raised {
    background-color: #252f49;
  }

  .incognito-mode .top-bar.raised:backdrop,
  .incognito-mode .bottom-bar.raised:backdrop {
    background-color: #1c2438;
  }
}

.incognito-mode .top-bar.raised {
  background-image: cross-fade(10% -gtk-icontheme('user-not-tracked-symbolic'), image(transparent));
  background-position: 168px calc((46px - 64px) / 2);
  background-size: 64px 64px;
  background-repeat: no-repeat;
}

.incognito-mode .top-bar.raised:backdrop {
  background-image: cross-fade(5% -gtk-icontheme('user-not-tracked-symbolic'), image(transparent));
  transition: background 200ms ease-out;
}

.incognito-mode.narrow .top-bar.raised {
  background-position: 16px calc((46px - 64px) / 2);
}

/* automation mode */
.automation-mode .top-bar.raised,
.automation-mode .bottom-bar.raised {
  background-color: #ffa348;
}

.automation-mode .top-bar.raised:backdrop,
.automation-mode .bottom-bar.raised:backdrop {
  background-color: #ffbe6f;
}

@media (prefers-color-scheme: dark) {
  /* dark style */
  .automation-mode .top-bar.raised,
  .automation-mode .bottom-bar.raised {
    background-color: #733100;
  }

  .automation-mode .top-bar.raised:backdrop,
  .automation-mode .bottom-bar.raised:backdrop {
    background-color: #632300;
  }
}

/* Fix awkward padding on the tabs dropdown arrow */
tabbar .end-action {
  padding: 0;
}

tabbar .end-action > revealer > menubutton {
  margin-top: 3px;
  margin-bottom: 3px;
}

tabbar .end-action > revealer > menubutton:dir(ltr) {
  margin-right: 5px;
}

tabbar .end-action > revealer > menubutton:dir(rtl) {
  margin-left: 5px;
}

/* location entry */
.location-entry {
  margin: 0;
}

.narrow .location-entry {
  font-size: small;
}

.entry-icon,
.entry-icon button
/* For the opensearch button because opensearch_button.entry_icon > menubutton > button
 * not menubutton.entry_icon > button like the other ones.
 */
.entry-icon menubutton {
  margin: 0;
}

.entry-icon > button {
  padding: 0;
  min-height: 0;
}

@media (prefers-contrast: more) {
  /* high contrast */
  .entry-icon {
    opacity: .85;
  }

  .entry-icon:hover {
    opacity: 1;
  }

  .entry-icon:active {
    opacity: .9;
  }
}

.entry-icon.checked > image {
  color: @accent_color;
}

.entry-icon.bookmarked > image {
  color: @yellow_5;
}

@media (prefers-color-scheme: dark) {
  /* dark style */
  .entry-icon.starred {
    color: @yellow_1;
  }
}

popover.suggestions {
  padding-top: 6px;

  /* Match entries */
  padding-left: 9px;
  padding-right: 9px;
}

popover.suggestions listview > row {
  padding: 0;
}

popover.suggestions listview > row > grid {
  padding: 9px 12px;
}

popover.suggestions listview > row:hover {
  background-color: alpha(currentColor, .04);
}

popover.suggestions listview > row:active {
  background-color: alpha(currentColor, .08);
}

popover.suggestions listview > row:selected {
  background-color: alpha(currentColor, .1);
}

popover.suggestions listview > row:selected:hover {
  background-color: alpha(currentColor, .13);
}

popover.suggestions listview > row:selected:active {
  background-color: alpha(currentColor, .19);
}

@media (prefers-contrast: more) {
  /* high contrast */
  popover.suggestions listview > row:selected {
    box-shadow: inset 0 0 0 1px @borders;
  }
}

/* bookmarks */

.bookmark-tag-widget {
  padding-left: 8px;
  background-color: alpha(currentColor, .1);
  border-radius: 6px;
}

.bookmark-tag-widget label {
  padding-left: 8px;
  padding-right: 8px;
}

.bookmark-tag-widget-selected {
  background-color: @accent_bg_color;
  color: @accent_fg_color;
}

.bookmark-tag-widget button {
  border-radius: 100%;
  min-width: 24px;
  min-height: 24px;
  padding: 0;
  margin: 2px;
}


/* "Press $KEY to exit fullscreen" overlay text */
#fullscreen-popup {
  padding: 12px;
  border-radius: 18px;
  background: rgba(0, 0, 0, 0.65);
  color: white;
}

dnd > .boxed-list {
  margin: 6px;
  background-color: @window_bg_color;
  background-image: image(@card_bg_color);
}

.message-popover > contents {
  padding: 0;
  margin: 0;
}

.message-popover .response-area > button {
  padding: 10px 14px;
  border-radius: 0;
  margin-top: -1px;
  margin-right: -1px;
  margin-left: -1px;
}

.message-popover .response-area > button.suggested {
  color: @accent_color;
}

.message-popover .response-area > button.destructive {
  color: @destructive_color;
}

.message-popover .response-area > button:first-child:dir(ltr),
.message-popover .response-area > button:last-child:dir(rtl) {
  border-bottom-left-radius: 12px;
  margin-left: 0;
}

.message-popover .response-area > button:last-child:dir(ltr),
.message-popover .response-area > button:first-child:dir(rtl) {
  border-bottom-right-radius: 12px;
  margin-right: 0;
}

@media (prefers-contrast: more) {
  /* high contrast */
  .response-area > button:hover,
  .response-area > button:active,
  .response-area > button:checked {
    box-shadow: none;
  }
}

.progress-indicator {
  margin-left: 6px;
  margin-right: 6px;
}
