/* floating status bar */
.floating-bar {
  background-color: rgba(51, 51, 51, 0.9);
  border: none;
  color: white;
  margin: 0.500000001rem;
  padding: 0.500000001rem 1.000000001rem;
  box-shadow: 0 1px 3px rgba(0, 0, 0, 0.12), 0 1px 2px rgba(0, 0, 0, 0.24);
  border-radius: 0.250000001rem;
}

/* header bar */

.toolbar-button > image,
.toolbar-button > button > box > image {
  -gtk-icon-size: 24px;
}

.page-menu-button-box {
  padding-left: 12px;
  padding-right: 12px;
  padding-bottom: 6px;
}

.page-menu-zoom-box {
  padding-top: 6px;
  padding-left: 12px;
  padding-right: 12px;
}

/* TODO: incognito */

/* location entry */
.location-entry {
  min-height: 24px;
}

.narrow .location-entry {
  font-size: small;
}

button.entry-icon,
menubutton.entry-icon > button {
  background: none;
}

button.entry-icon > image,
menubutton.entry-icon > button > box > image {
  padding: 0;
}

popover.suggestions {
  padding-top: 0.500000001rem;
}

popover.suggestions contents {
  padding: 0;
}

popover.suggestions contents listview {
  padding-top: 0.500000001rem;
  padding-bottom: 0.500000001rem;
}

popover.suggestions listview > row {
  padding: 0;
}

popover.suggestions listview > row > grid {
  padding: 0.500000001rem 1.000000001rem;
}

popover.suggestions listview > row:hover,
popover.suggestions listview > row:selected {
  background-color: alpha(currentColor, 0.15);
}

/* bookmarks */
.bookmarks-row:hover {
  background: alpha(currentColor, 0.15);
}

.bookmarks-row button {
  opacity: 0.25;
}

.bookmarks-row:hover button {
  opacity: 1;
}

.bookmark-tag-widget {
  padding-left: 8px;
  background-color: alpha(currentColor, .1);
  border-radius: 6px;
}

.bookmark-tag-widget label {
  padding-left: 8px;
  padding-right: 8px;
}

.bookmark-tag-widget-selected {
  background-color: @theme_selected_bg_color;
  color: @theme_selected_fg_color;
}

.bookmark-tag-widget button {
  border-radius: 100%;
  min-width: 24px;
  min-height: 24px;
  padding: 0;
  margin: 2px;
}

/* "Press $KEY to exit fullscreen" overlay text */
#fullscreen-popup {
  padding: 12px;
  border-radius: 18px;
  background: rgba(0, 0, 0, 0.65);
  color: white;
}
